"""
These functions all accept, as additional keyword arguments, any arguments
of ``Config``.
"""

from __future__ import absolute_import

__all__ = []

import functools
import json

from . import urls
from .config import *
from .exceptions import *
from .map_list import *
from .models import *

base_url = 'https://api.chain.com'

make_url = functools.partial(urls.make_url, base_url)

json_header = {'content-type': 'application/json'}


def json_args(x):
    return dict(data=json.dumps(x), headers=json_header)


def parse_response_data(response):
    x = response.json()
    if 'error' in x:
        raise ChainError(x['error'])
    return x


def endpoint(f):
    __all__.append(f.__name__)
    return f


@endpoint
def get_address(address, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-address

    Arguments:

        address: string
        A Bitcoin address.

    Return: ``Address``
    """
    config = config.replace(**kw)
    require_api_key_id(config)
    requests = config_to_requests(config)
    url = make_url(['v1', config.block_chain, 'addresses', address])
    response_data = parse_response_data(requests.get(url))
    return Address.from_dict(response_data)


@endpoint
def get_addresses(addresses, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-address

    Arguments:

        addresses: list of string
        A set of Bitcoin addresses.

    Return: list of ``Address``
    """
    config = config.replace(**kw)
    require_api_key_id(config)
    requests = config_to_requests(config)
    url = make_url(
        ['v1', config.block_chain, 'addresses', ','.join(addresses)]
    )
    response_data = parse_response_data(requests.get(url))
    return map_list(Address.from_dict, response_data)


@endpoint
def get_address_transactions(address, limit=None, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-address-transactions

    Arguments:

        address: string
        A Bitcoin address.

        limit: number
        The max number of transactions to return, starting with most recent.

    Return: list of ``Transaction``
    """
    config = config.replace(**kw)
    require_api_key_id(config)
    requests = config_to_requests(config)
    url = make_url(
        ['v1', config.block_chain, 'addresses', address, 'transactions'],
        [('limit', limit)]
    )
    response_data = parse_response_data(requests.get(url))
    return map_list(Transaction.from_dict, response_data)


@endpoint
def get_addresses_transactions(addresses, limit=None, config=default_config,
                               **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-address-transactions

    Arguments:

        addresses: list of string
        A set of Bitcoin addresses.

        limit: number
        The max number of transactions to return, starting with most recent.

    Return: list of ``Transaction``
    """
    config = config.replace(**kw)
    require_api_key_id(config)
    requests = config_to_requests(config)
    url = make_url(
        ['v1', config.block_chain, 'addresses', ','.join(addresses),
         'transactions'],
        [('limit', limit)]
    )
    response_data = parse_response_data(requests.get(url))
    return map_list(Transaction.from_dict, response_data)


@endpoint
def get_address_unspents(address, limit=None, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-address-unspents

    Arguments:

        address: string
        A Bitcoin address.

        limit: number
        The max number of transactions to return, starting with most recent.

    Return: list of ``Output``
    """
    config = config.replace(**kw)
    require_api_key_id(config)
    requests = config_to_requests(config)
    url = make_url(
        ['v1', config.block_chain, 'addresses', address, 'unspents'],
        [('limit', limit)]
    )
    response_data = parse_response_data(requests.get(url))
    return map_list(Output.from_dict, response_data)


@endpoint
def get_addresses_unspents(addresses, limit=None, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-address-unspents

        addresses: list of string
        A set of Bitcoin addresses.

        limit: number
        The max number of transactions to return, starting with most recent.

    Return: list of ``Output``
    """
    config = config.replace(**kw)
    require_api_key_id(config)
    requests = config_to_requests(config)
    url = make_url(
        ['v1', config.block_chain, 'addresses', ','.join(addresses),
         'unspents'],
        [('limit', limit)]
    )
    response_data = parse_response_data(requests.get(url))
    return map_list(Output.from_dict, response_data)


@endpoint
def get_address_op_returns(address, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-address-op-returns

    Arguments:

        address: string
        A Bitcoin address.

    Return: list of ``OpReturn``
    """
    config = config.replace(**kw)
    require_api_key_id(config)
    requests = config_to_requests(config)
    url = make_url(
        ['v1', config.block_chain, 'addresses', address, 'op-returns']
    )
    response_data = parse_response_data(requests.get(url))
    return map_list(OpReturn.from_dict, response_data)


@endpoint
def get_transaction(transaction_hash, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-transaction

    Arguments:

        hash: string
        A transaction hash.

    Return: ``Transaction``
    """
    config = config.replace(**kw)
    require_api_key_id(config)
    requests = config_to_requests(config)
    url = make_url(['v1', config.block_chain, 'transactions', transaction_hash])
    response_data = parse_response_data(requests.get(url))
    return Transaction.from_dict(response_data)


@endpoint
def get_transaction_op_return(transaction_hash, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-transaction-op-return

    Arguments:

        hash: string
        A transaction hash.

    Return: ``OpReturn``
    """
    config = config.replace(**kw)
    require_api_key_id(config)
    requests = config_to_requests(config)
    url = make_url(['v1', config.block_chain, 'transactions', transaction_hash,
                    'op-return'])
    response_data = parse_response_data(requests.get(url))
    return OpReturn.from_dict(response_data)


@endpoint
def send_transaction(hex, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-transaction-send

    Arguments:

        hex: string
        A hex representation of the signed transaction.

    Return: ``SendTransactionResult``
    """
    config = config.replace(**kw)
    require_api_key_secret(config)
    requests = config_to_requests(config)
    url = make_url(['v1', config.block_chain, 'transactions'])
    request_data = {'hex': hex}
    response = requests.put(url, **json_args(request_data))
    response_data = parse_response_data(response)
    return SendTransactionResult.from_dict(response_data)


def _get_block(x, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-block

    Return: ``Block``
    """
    config = config.replace(**kw)
    require_api_key_id(config)
    requests = config_to_requests(config)
    url = make_url(['v1', config.block_chain, 'blocks', x])
    response_data = parse_response_data(requests.get(url))
    return Block.from_dict(response_data)


@endpoint
def get_block_by_hash(block_hash, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-block

    Arguments:

        hash: string
        A block hash.

    Return: ``Block``
    """
    return _get_block(block_hash, config, **kw)


@endpoint
def get_block_by_height(height, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-block

    Arguments:

        height: integer
        A block height.

    Return: ``Block``
    """
    return _get_block(str(height), config, **kw)


@endpoint
def get_latest_block(config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-block

    Return: ``Block``
    """
    return _get_block('latest', config, **kw)


def _get_block_op_returns(x, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-block-op-returns

    Return: ``OpReturn``
    """
    config = config.replace(**kw)
    require_api_key_id(config)
    requests = config_to_requests(config)
    url = make_url(['v1', config.block_chain, 'blocks', x, 'op-returns'])
    response_data = parse_response_data(requests.get(url))
    return map_list(OpReturn.from_dict, response_data)


@endpoint
def get_block_op_returns_by_hash(block_hash, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-block-op-returns

    Arguments:

        hash: string
        A block hash.

    Return: ``OpReturn``
    """
    return _get_block_op_returns(block_hash, config, **kw)


@endpoint
def get_block_op_returns_by_height(height, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#bitcoin-block-op-returns

    Arguments:

        height: integer
        A block height.

    Return: ``OpReturn``
    """
    return _get_block_op_returns(str(height), config, **kw)


@endpoint
def create_webhook(webhook_url, webhook_id=None, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#webhooks-create

    Arguments:

        webhook_url: string
        The URL of the server that will receive the Webhook POST requests.

        webhook_id: string
        A user-generated id for the Webhook. If you do not provide an id,
        one will be automatically generated by the API.

    Return: ``Webhook``
    """
    config = config.replace(**kw)
    require_api_key_secret(config)
    requests = config_to_requests(config)
    url = make_url(['v1', 'webhooks'])
    request_data = {'id': webhook_id, 'url': webhook_url}
    response = requests.post(url, **json_args(request_data))
    response_data = parse_response_data(response)
    return Webhook.from_dict(response_data)


@endpoint
def list_webhooks(config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#webhooks-list

    Return: list of ``Webhook``
    """
    config = config.replace(**kw)
    require_api_key_secret(config)
    requests = config_to_requests(config)
    url = make_url(['v1', 'webhooks'])
    response_data = parse_response_data(requests.get(url))
    return map_list(Webhook.from_dict, response_data)


@endpoint
def update_webhook(webhook_url, webhook_id, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#webhooks-update

    Arguments:

        webhook_url: string
        The new server URL that will receive the Webhook POST requests.

        webhook_id: string
        The unique id of the Webhook.

    Return: ``Webhook``
    """
    config = config.replace(**kw)
    require_api_key_secret(config)
    requests = config_to_requests(config)
    url = make_url(['v1', 'webhooks', webhook_id])
    request_data = {'url': webhook_url}
    response = requests.put(url, **json_args(request_data))
    response_data = parse_response_data(response)
    return Webhook.from_dict(response_data)


@endpoint
def delete_webhook(webhook_id, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#webhooks-delete

    Arguments:

        id: string
        The unique id of the Webhook.

    Return: ``Webhook``
    """
    config = config.replace(**kw)
    require_api_key_secret(config)
    requests = config_to_requests(config)
    url = make_url(['v1', 'webhooks', webhook_id])
    response = requests.delete(url)
    response_data = parse_response_data(response)
    return Webhook.from_dict(response_data)


@endpoint
def create_address_transaction_event(webhook_id, address, confirmations=None,
                                     config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#webhooks-event-create

    Arguments:

        webhook_id: string - The unique id of the associated Webhook.

        address: string - The address that will be used to match Webhook
        Events.

        confirmations: number - The number of confirmations that will be
        POSTed to the Webhook for each new transaction.

    Return: ``AddressTransactionEvent``
    """
    config = config.replace(**kw)
    require_api_key_secret(config)
    requests = config_to_requests(config)
    url = make_url(['v1', 'webhooks', webhook_id, 'events'])
    request_data = {
        'event': AddressTransactionEvent.event,
        'block_chain': config.block_chain,
        'address': address,
    }
    if confirmations is not None:
        request_data['confirmations'] = confirmations
    response = requests.post(url, **json_args(request_data))
    response_data = parse_response_data(response)
    return AddressTransactionEvent.from_dict(response_data)


@endpoint
def list_webhook_events(webhook_id, config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#webhooks-event-list

    Arguments:

        webhook_id: string
        The unique id of the associated Webhook.

    Return: list of ``WebhookEvent``
    """
    config = config.replace(**kw)
    require_api_key_secret(config)
    requests = config_to_requests(config)
    url = make_url(['v1', 'webhooks', webhook_id, 'events'])
    response_data = parse_response_data(requests.get(url))
    return map_list(WebhookEvent.from_dict, response_data)


@endpoint
def delete_address_transaction_event(webhook_id, address,
                                     config=default_config, **kw):
    """
    https://chain.com/docs/v1/curl/#webhooks-event-delete

    Arguments:

        webhook_id: string
        The unique id of the associated Webhook.

        address: string
        The address that is used to match Webhook Events.

    Return: ``AddressTransactionEvent``
    """
    config = config.replace(**kw)
    require_api_key_secret(config)
    requests = config_to_requests(config)
    url = make_url(['v1', 'webhooks', webhook_id, 'events',
                    AddressTransactionEvent.event, address])
    response_data = parse_response_data(requests.delete(url))
    return AddressTransactionEvent.from_dict(response_data)



from __future__ import absolute_import

__all__ = ('make_url', 'make_query_string')

try:
    from urllib.parse import urlencode  # python 3
except ImportError:
    from urllib import urlencode  # python 2

import re

path_component_regex = re.compile(r'^[0-9a-z_\-,]+$', re.I)


def make_url(base, path_components, query_params=None):

    for c in path_components:
        if not path_component_regex.match(c):
            raise ValueError('Invalid character in "{0}"'.format(c))

    return '?'.join(filter(None, [
        '/'.join([base] + path_components),
        make_query_string(query_params),
    ]))


def make_query_string(query_params):
    query_params = filter(lambda p: p[1] is not None, query_params or [])
    return '&'.join(map(url_encode_tuple, query_params))


def url_encode_tuple(tup):
    return urlencode(dict([tup]))



from __future__ import absolute_import

__all__ = (
    'namedtuple', 'namedtuple_to_dict', 'transform_item', 'remove_item'
)

import logging
logger = logging.getLogger('chain_bitcoin')

import collections
import functools
import json
import types

from .exceptions import *
from .func_util import set_function_defaults


def namedtuple(name, args=None, alter_dict=None):
    """
    A class constructor similar to ``collections.namedtuple``.

    Arguments:

        args: list of strings

        alter_dict: map from dict to dict
        Modifies the behavior of ``from_dict``.

    Differences from ``collections.namedtuple``:

        All parameters to ``__new__`` have a default value of ``None``.

        ``_replace`` is aliased as ``replace`` because it shouldn't be
        considered a private method.

        There is a class method ``from_dict`` which converts a ``dict``
        to an instance of the ``namedtuple`` class.

        ``args`` must be a list of strings.
    """

    args = list(args or [])

    tuple_type = collections.namedtuple(name, args)

    set_function_defaults(tuple_type.__new__, [None] * len(args))

    tuple_type.replace = tuple_type._replace

    def from_dict(cls, x):
        return namedtuple_from_dict(cls, alter_dict(x) if alter_dict else x)
    tuple_type.from_dict = classmethod(from_dict)

    return tuple_type


def namedtuple_from_dict(tuple_type, x):

    # Remove keys that aren't fields, and log a warning.
    bad_keys = list(k for k in x if k not in tuple_type._fields)
    if len(bad_keys) != 0:
        logger.warning('Keys not applicable to {name}: {keys}\n{x}'.format(
            name=tuple_type.__name__,
            keys=', '.join(bad_keys),
            x=str(x),
        ))
        x = dict((k, x[k]) for k in x if k not in bad_keys)

    return tuple_type(**x)


def transform_item(k, f):
    def t(x):
        x = dict(x)
        if k in x:
            x[k] = f(x[k])
        return x
    return t


def remove_item(k):
    def t(x):
        if k in x:
            x = dict(x)
            del x[k]
        return x
    return t


def namedtuple_to_dict(tup):
    return dict((s, getattr(tup, s)) for s in tup._fields)



from __future__ import absolute_import

__all__ = ('Chain',)

import inspect
import textwrap

from . import endpoints
from .config import *
from .exceptions import *
from .func_util import *
from .models import *
from .namedtuple import *


class Chain(object):

    def __init__(self, config=default_config, **kw):
        """
        Keyword arguments: Any arguments of ``Config``.
        """
        self.config = config.replace(**kw)

    def configure(self, **kw):
        """
        Keyword arguments: Any arguments of ``Config``.

        Returns a new ``Chain``.
        """
        return Chain(self.config, **kw)

    @property
    def requests(self):
        """
        Returns a ``requests.Session``.
        """
        return config_to_requests(self.config)


# Copy functions from the ``endpoints`` module onto the ``Chain`` class.
for name in endpoints.__all__:
    f = getattr(endpoints, name)
    argspec = inspect.getargspec(f)

    # Remove the ``config`` param
    argspec = argspec_drop_right(1, argspec)

    # Add a ``self`` param
    argspec = argspec_append_left(['self'], argspec)

    code = textwrap.dedent("""
        def _{name}{argspec}:
            return endpoints.{name}({args}config=self.config, **kw)
        Chain.{name} = _{name}
    """).format(
        argspec=inspect.formatargspec(*argspec),
        args=''.join(map('{0}, '.format, argspec.args[1:])),
        name=name,
    )
    exec(code, globals())



from __future__ import absolute_import

__all__ = (
    'set_function_defaults', 'argspec_drop_right', 'argspec_append_left'
)

from inspect import ArgSpec


def set_function_defaults(f, defaults):
    if hasattr(f, '__defaults__'):  # python 3
        f.__defaults__ = tuple(defaults) or None
    elif hasattr(f, 'func_defaults'):  # python 2
        f.func_defaults = tuple(defaults) or None
    else:
        raise Exception('???')


def argspec_drop_right(n, argspec):
    return argspec._replace(args=argspec.args[:-n],
                            defaults=argspec.defaults[:-n] or None)


def argspec_append_left(new_args, argspec):
    return argspec._replace(args=(new_args + argspec.args))


def compose(*fs):
    """
    compose(a, b, c)(x) = a(b(c(x)))
    """
    def g(x):
        for f in reversed(fs):
            x = f(x)
        return x

    return g



from __future__ import absolute_import

__all__ = ('ChainError', 'NoApiKeyId', 'NoApiKeySecret')


class ChainError(BaseException):
    pass


class NoApiKeyId(ChainError):
    pass


class NoApiKeySecret(ChainError):
    pass



from __future__ import absolute_import

__all__ = ('Config', 'ca_cert', 'default_block_chain', 'default_config',
           'config_to_requests', 'require_api_key_id', 'require_api_key_secret')

import pkg_resources
import requests
import requests.auth

from .exceptions import *
from .namedtuple import namedtuple


class Config(namedtuple('Config', (
    'api_key_id', 'api_key_secret', 'block_chain', 'verify', 'http_adapter'
))):
    """
    Arguments:

        api_key_id: string

        api_key_secret: string

        block_chain: string - The name of the block chain.
        For example: bitcoin or testnet3.
        https://chain.com/docs/v1/curl/#block-chains

        verify - One of:
        - A CA_BUNDLE path for SSL cert verification,
        - ``True`` to verify using the system CA_BUNDLE
        - ``False`` to skip SSL cert verification.
        The default is verification with Chain's CA cert.

        http_adapter: ``requests.adapters.HTTPAdapter``
        The HTTP adapter used to make HTTP requests.
        (This parameter mostly exists for testing.)
    """


ca_cert = pkg_resources.resource_filename(__name__, 'chain.pem')

default_block_chain = 'bitcoin'

default_config = Config(verify=ca_cert, block_chain=default_block_chain)


def config_to_requests(config):

    session = requests.Session()
    session.verify = config.verify

    if config.api_key_id is not None:

        if config.api_key_secret is not None:
            session.auth = requests.auth.HTTPBasicAuth(
                config.api_key_id, config.api_key_secret)

        else:
            session.params.update({'api-key-id': config.api_key_id})

    if config.http_adapter is not None:
        session.mount('http://', config.http_adapter)
        session.mount('https://', config.http_adapter)

    return session


def require_api_key_id(config):
    if config.api_key_id is None:
        raise NoApiKeyId


def require_api_key_secret(config):
    require_api_key_id(config)
    if config.api_key_secret is None:
        raise NoApiKeySecret



from __future__ import absolute_import

__all__ = ('map_list',)

from .exceptions import *


def map_list(f, xs):
    """
    Use ``f`` to transform each element of ``xs`` if ``xs`` is a list.
    Raise ``ChainError`` if ``xs`` is not a list.
    """
    if not isinstance(xs, list):
        raise ChainError('Expected list, got "{0}"'.format(xs))
    return list(map(f, xs))



from setuptools import setup
import pkg_resources
import re


def read(path):
    with open(pkg_resources.resource_filename(__name__, path)) as f:
        return f.read()

def long_description():
    return re.split('\n\.\. pypi [^\n]*\n', read('README.rst'), 1)[1]


setup(
    name='chain_bitcoin',
    version='0.3',
    author='Chris Martin',
    author_email='ch.martin@gmail.com',
    packages=['chain_bitcoin'],
    package_data={'chain_bitcoin': ['chain.pem']},
    include_package_data=True,
    url='https://github.com/cardforcoin/chain-bitcoin-python',
    license='MIT',
    description='Integration library for the Chain.com API',
    long_description=long_description(),
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    install_requires=read('requirements-install.txt').split('\n'),
    tests_require=read('requirements-test.txt').split('\n'),
)



from __future__ import absolute_import

__all__ = (
    'Address', 'Transaction', 'Output', 'OpReturn', 'Block', 'Webhook',
    'WebhookEvent', 'EchoVerificationEvent', 'AddressTransactionEvent',
    'SendTransactionResult'
)

from enum import Enum

from .func_util import compose
from .map_list import *
from .namedtuple import *


class Address(namedtuple('Address', (
    'hash', 'balance', 'received', 'sent', 'unconfirmed_received',
    'unconfirmed_sent', 'unconfirmed_balance'
))):
    """
    https://chain.com/docs/v1/curl/#object-bitcoin-address

    The Address Object contains basic balance details for a Bitcoin address.
    """


class Transaction(namedtuple('Transaction', (
    'hash', 'block_hash', 'block_height', 'block_time', 'inputs', 'outputs',
    'amount', 'fees', 'confirmations', 'chain_received_at', 'propagation_level',
    'double_spend'
), alter_dict=compose(
    transform_item(
        'inputs', lambda v: map_list(Transaction.Input.from_dict, v)),
    transform_item(
        'outputs', lambda v: map_list(Transaction.Output.from_dict, v)),
))):
    """
    https://chain.com/docs/v1/curl/#object-bitcoin-transaction

    The Transaction Object contains details about a Bitcoin transaction,
    including inputs and outputs.
    """

    class Input(namedtuple('Transaction_Input', (
        'transaction_hash', 'output_hash', 'output_index', 'value', 'addresses',
        'script_signature', 'coinbase'
    ))):
        """
        https://chain.com/docs/v1/curl/#object-bitcoin-transaction

        Values of ``Transaction.inputs``.
        """

    class Output(namedtuple('Transaction_Output', (
        'transaction_hash', 'output_index', 'value', 'addresses', 'script',
        'script_hex', 'script_type', 'required_signatures', 'spent',
    ))):
        """
        https://chain.com/docs/v1/curl/#object-bitcoin-transaction

        Values of ``Transaction.outputs``.
        """


class SendTransactionResult(namedtuple('SendTransactionResult', ('hash',))):
    """
    https://chain.com/docs/v1/curl/#bitcoin-transaction-send

    The value returned by the ``send_transaction`` endpoint.

    Arguments:

        hash: string - The newly created transaction hash.
    """


class Output(namedtuple('TransactionOutput', (
    'transaction_hash', 'output_index', 'value', 'addresses', 'script',
    'script_hex', 'script_type', 'required_signatures', 'spent',
    'confirmations'
))):
    """
    https://chain.com/docs/v1/curl/#bitcoin-address-unspents

    The Output Object is a pseudo-object that is extracted from the
    Transaction Object for use when the entire Transaction Object is not
    needed.
    """


class OpReturn(namedtuple('OpReturn', (
    'transaction_hash', 'hex', 'text', 'sender_addresses', 'receiver_addresses'
))):
    """
    https://chain.com/docs/v1/curl/#bitcoin-address-op-returns

    The OP_RETURN Object is a pseudo-object that is extracted from the
    Transaction Object. It is an interpretation of the OP_RETURN script
    within a zero-value output in a Bitcoin transaction. The OP_RETURN can
    be used to include 40 bytes of metadata in a Bitcoin transaction.
    """


class Block(namedtuple('Block', (
    'hash', 'previous_block_hash', 'height', 'confirmations', 'merkle_root',
    'time', 'nonce', 'difficulty', 'transaction_hashes'
))):
    """
    https://chain.com/docs/v1/curl/#object-bitcoin-block

    The Block Object contains details about a Bitcoin block, including all
    transaction hashes.
    """


class Webhook(namedtuple('Webhook', ('id', 'url'))):
    """
    https://chain.com/docs/v1/curl/#object-webhooks

    The Webhook Object contains a server URL that the Chain API uses to
    communicate with your application. You can create one or more Webhook
    Objects, which may each have one or more associated Event Objects.
    """


class WebhookEvent:
    """
    https://chain.com/docs/v1/curl/#object-webhooks-event
    """

    event = None
    """
    https://chain.com/docs/v1/curl/#object-webhooks-event

    The event that will trigger the Webhook's POST request.
    """

    class Message(object):
        """
        https://chain.com/docs/v1/curl/#webhooks-receiving

        The class for data sent to the webhook URL.
        """

        @classmethod
        def from_dict(cls, x):
            x = dict(x)
            event_type = next(t for t in webhook_event_types
                              if t.event == x['event'])
            return event_type.Message.from_dict(x)

    @classmethod
    def from_dict(cls, x):
        x = dict(x)
        event_type = next(t for t in webhook_event_types
                          if t.event == x['event'])
        return event_type.from_dict(x)


echo_verification_event = 'echo-verification'

address_transaction_event = 'address-transaction'


class EchoVerificationEvent(
    namedtuple('EchoVerificationEvent', alter_dict=remove_item('event'))
):
    """
    https://chain.com/docs/v1/curl/#webhooks-setup

    Each time you create a new Webhook, before completing the request, the
    Chain API will make a test request to the Webhook in order to verify
    ownership.

    To pass the verification test, your web server must respond with an exact
    copy of the request body.
    """

    event = echo_verification_event

    class Message(namedtuple(
        'EchoVerificationEvent_Message',
        alter_dict=remove_item('event')
    )):
        pass


class AddressTransactionEvent(
    namedtuple('AddressTransactionEvent', (
        'id', 'webhook_id', 'block_chain', 'address', 'confirmations'
    ), alter_dict=remove_item('event')),
    WebhookEvent,
):
    """
    https://chain.com/docs/v1/curl/#object-webhooks-event

    This event triggers when a new transaction occurs on a specified address.
    The first POST will notify your application of the new, unconfirmed
    transaction. Additional POSTs will notify your application of subsequent
    confirmations for that transction.

    Arguments:

        id: string

        webhook_id: string - The unique id of the associated Webhook.

        block_chain: string - The name of the block chain that the Webhook
        Event is associated with.

        address: string - The address that will be used to match Webhook
        Events.

        confirmations: number - The number of confirmations that will be
        POSTed to the Webhook for each new transaction.
    """

    event = address_transaction_event

    class Message(namedtuple('AddressTransactionEvent_Message', (
        'transaction', 'block_chain', 'address'
    ), alter_dict=compose(
        transform_item('transaction', Transaction.from_dict),
        remove_item('event'),
    ))):
        """
        https://chain.com/docs/v1/curl/#webhooks-receiving

        The data structure that will be POSTed to your server from an
        address-transaction event.
        """

        event = address_transaction_event


webhook_event_types = (
    EchoVerificationEvent,
    AddressTransactionEvent,
)
